import Client as pe_
import os
import random
import numpy as np
import copy
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.metrics import mean_squared_error
from matplotlib import pyplot
import pickle
from sklearn.externals import joblib
import time

DISCARD_DATA = '../data/discard-data.sav'

card_val = {
    '2h': 1,
    '3h': 2,
    '4h': 3,
    '5h': 4,
    '6h': 5,
    '7h': 6,
    '8h': 7,
    '9h': 8,
    'Th': 9,
    'Jh': 10,
    'Qh': 11,
    'Kh': 12,
    'Ah': 13,
    '2d': 14,
    '3d': 15,
    '4d': 16,
    '5d': 17,
    '6d': 18,
    '7d': 19,
    '8d': 20,
    '9d': 21,
    'Td': 22,
    'Jd': 23,
    'Qd': 24,
    'Kd': 25,
    'Ad': 26,
    '2s': 27,
    '3s': 28,
    '4s': 29,
    '5s': 30,
    '6s': 31,
    '7s': 32,
    '8s': 33,
    '9s': 34,
    'Ts': 35,
    'Js': 36,
    'Qs': 37,
    'Ks': 38,
    'As': 39,
    '2c': 40,
    '3c': 41,
    '4c': 42,
    '5c': 43,
    '6c': 44,
    '7c': 45,
    '8c': 46,
    '9c': 47,
    'Tc': 48,
    'Jc': 49,
    'Qc': 50,
    'Kc': 51,
    'Ac': 52
 }


"""
[1,0,0,0,0,0,0,0,0,0] SF
[0,1,0,0,0,0,0,0,0,0] 4OfAKind
[0,0,1,0,0,0,0,0,0,0] FullHouse
[0,0,0,1,0,0,0,0,0,0] Flush
[0,0,0,0,1,0,0,0,0,0] Straight
[0,0,0,0,0,1,0,0,0,0] 3OfAkind
[0,0,0,0,0,0,1,0,0,0] TwoPairs
[0,0,0,0,0,0,0,1,0,0] OnePair
[0,0,0,0,0,0,0,0,1,0] HighCard > 9
[0,0,0,0,0,0,0,0,0,1] HighCard <= 9
"""

HandValue = {
    'HighCard_under':[0,0,0,0,0,0,0,0,0,1],
    'HighCard_over': [0,0,0,0,0,0,0,0,1,0],
    'OnePair':       [0,0,0,0,0,0,0,1,0,0],
    'TwoPairs':      [0,0,0,0,0,0,1,0,0,0],
    '3OfAKind':      [0,0,0,0,0,1,0,0,0,0],
    'Straight':      [0,0,0,0,1,0,0,0,0,0],
    'Flush':         [0,0,0,1,0,0,0,0,0,0],
    'FullHouse':     [0,0,1,0,0,0,0,0,0,0],
    '4OfAKind':      [0,1,0,0,0,0,0,0,0,0],
    'StraightFlush': [1,0,0,0,0,0,0,0,0,0]
}

RemovedCard = {
    0: [0,0,0,0,0,1],
    1: [0,0,0,0,1,0],
    2: [0,0,0,1,0,0],
    3: [0,0,1,0,0,0],
    4: [0,1,0,0,0,0],
    5: [1,0,0,0,0,0]
}

ChangeAction = {
    0: "Save card",
    1: "Change 1",
    2: "Change 2",
    3: "Change 3",
    4: "Change 4",
    5: "Change whole hand!"
}

def card_sorter(Hand_):

    # Get the type of Hand
    def ev(Hand_):
        count = 0
        for card1 in Hand_:
            for card2 in Hand_:
                if (card1[0] == card2[0]) and (card1[1] != card2[1]):
                    count += 1
        return count

    # Use the "count" to analyse hand
    count_ = ev(Hand_)

    sub1 = 0
    score = [' ', ' ', ' ', ' ', ' ', ' ']
    check_card = []
    rest = []

    if count_ == 12:
        for card1 in Hand_:
            for card2 in Hand_:
                if (card1[0] == card2[0]) and (card1[1] != card2[1]):
                    sub1 += 1
                    if card1 not in check_card:
                        check_card.append(card1)
                    check_card.append(card2)

            if sub1 == 3:
                for card_ in Hand_:
                    if card_ not in check_card:
                        rest.append(card_)
                quicksort(rest)
                quicksort(check_card)
                score = ['4OfAKind', rest[0], check_card[0], check_card[1], check_card[2], check_card[3]]
                break

    elif count_ == 8:
        for card1 in Hand_:
            for card2 in Hand_:
                if (card1[0] == card2[0]) and (card1[1] != card2[1]):
                    sub1 += 1
            if sub1 == 1:
                sub1 = 0
            if sub1 == 2:
                quicksort(Hand_)
                score = ['FullHouse', Hand_[0], Hand_[1], Hand_[2], Hand_[3], Hand_[4]]
                break

    elif count_ == 6:
        for card1 in Hand_:
            for card2 in Hand_:
                if (card1[0] == card2[0]) and (card1[1] != card2[1]):
                    sub1 += 1
                    if card1 not in check_card:
                        check_card.append(card1)
                    check_card.append(card2)
            if sub1 == 2:
                for card_ in Hand_:
                    if card_ not in check_card:
                        rest.append(card_)
                quicksort(rest)
                score = ['3OfAKind', rest[0], rest[1], check_card[0], check_card[1], check_card[2]]
                break

    elif count_ == 4:
        needCard1 = ['', '']
        needCard2 = ['', '']
        for card1 in Hand_:
            for card2 in Hand_:
                # card1 keep the first hand card, card1 use every card to compare the card1
                if card1[0] == card2[0] and card1[1] != card2[1]:
                    if pe_.Suits[card1[1]] > pe_.Suits[card2[1]]:
                        if needCard1 == ['', '']:
                            needCard1 = card1
                    else:
                        if needCard1 == ['', '']:
                            needCard1 = card2
                if card1[0] == card2[0] and card1[1] != card2[1] \
                        and card1[0] != needCard1[0] and card2[0] != needCard1[0]:
                    if pe_.Suits[card1[1]] > pe_.Suits[card2[1]]:
                        if needCard2 == ['', '']:
                            needCard2 = card1
                    else:
                        if needCard2 == ['', '']:
                            needCard2 = card2
        twop = []
        for r in Hand_:
            if r[0] is not needCard1[0] and r[0] is not needCard2[0]:
                rest.append(r)
            elif r[0] is needCard1[0]:
                twop.append(r)
            else:
                twop.append(r)
        quicksort(twop)
        score = ['TwoPairs', rest[0], twop[0], twop[1], twop[2], twop[3]]

    elif count_ == 2:
        for card1 in Hand_:
            for card2 in Hand_:
                if (card1[0] == card2[0]) and (card1[1] > card2[1]):
                    sub1 += 1
                    if card1 not in check_card:
                        check_card.append(card1)
                    check_card.append(card2)
            if sub1 == 1:
                for card_ in Hand_:
                    if card_ not in check_card:
                        rest.append(card_)
                quicksort(rest)

                score = ['OnePair', rest[0], rest[1], rest[2], check_card[0], check_card[1]]
                break

    elif count_ == 0:

        quicksort(Hand_)
        score = ['HighCard', Hand_[0], Hand_[1], Hand_[2], Hand_[3], Hand_[4]]

        if Hand_[0][1] == Hand_[1][1] == Hand_[2][1] == Hand_[3][1] == Hand_[4][1]:
            score = ['Flush', Hand_[0], Hand_[1], Hand_[2], Hand_[3], Hand_[4]]

        if (pe_.Ranks[Hand_[4][0]] - pe_.Ranks[Hand_[3][0]] == 1) \
                and (pe_.Ranks[Hand_[3][0]] - pe_.Ranks[Hand_[2][0]] == 1) \
                and (pe_.Ranks[Hand_[2][0]] - pe_.Ranks[Hand_[1][0]] == 1) \
                and (pe_.Ranks[Hand_[1][0]] - pe_.Ranks[Hand_[0][0]] == 1):
            score = ['Straight', Hand_[0], Hand_[1], Hand_[2], Hand_[3], Hand_[4]]

            if Hand_[0][1] == Hand_[1][1] == Hand_[2][1] == Hand_[3][1] == Hand_[4][1]:
                score = ['StraightFlush', Hand_[0], Hand_[1], Hand_[2], Hand_[3], Hand_[4]]
    else:
        exit(5664)
    return score

def evaluate_hand(_hand):
    return pe_.Types[_hand[0]]*len(pe_.Ranks)+pe_.Ranks[_hand[4][0]]

def quicksort(items):

    def sort(lst, l, r):
        # base case
        if r <= l:
            return

        # choose random pivot
        pivot_index = random.randint(l, r)

        # move pivot to first index
        lst[l], lst[pivot_index] = lst[pivot_index], lst[l]

        # partition
        i = l
        for j in xrange(l+1, r+1):
            if pe_.Ranks[lst[j][0]] < pe_.Ranks[lst[l][0]]:
                i += 1
                lst[i], lst[j] = lst[j], lst[i]

        # place pivot in proper position
        lst[i], lst[l] = lst[l], lst[i]

        # sort left and right partitions
        sort(lst, l, i-1)
        sort(lst, i+1, r)

    if items is None or len(items) < 2:
        return

    sort(items, 0, len(items) - 1)

def serialize(_hand):
    if _hand[0] == 'HighCard':
        if pe_.Ranks[_hand[4][0]] > 9:
            _hand[0] = 'HighCard_over'
        else:
            _hand[0] = 'HighCard_under'
    return HandValue[_hand[0]]

class discard_ai(object):
    def __init__(self, data=DISCARD_DATA):
        self.model = joblib.load(DISCARD_DATA)
        if self.model == None:
            print "ERROR WITH LOADING DATA"

    def predict(self, _hand):
        _hand_strength = evaluate_hand(_hand)
        _serialized = serialize(_hand)
        _ser_hs = [i * _hand_strength for i in _serialized]
        allFeatures = []
        for a_ in range(0,6): allFeatures.append(_ser_hs+RemovedCard[a_])
        pReturn = self.model.predict(allFeatures)
        #action = ChangeAction[np.argmax(pReturn)]
        # RETURNING HIGHEST PREDICTED
        return pReturn

coolstuff = {
    0: '|',
    1: '/',
    2: '-',
    3: '\\'
}

def gen_dataset(num_samples=None):
    X,Y = [], []
    funny = 0
    mul = 1
    count = 0
    for x_ in range(0,num_samples):
        _hand = pe_.generate_hand()
        _hand = card_sorter(_hand)
        _start_hand = copy.deepcopy(_hand)
        hand_strengthx = evaluate_hand(_start_hand)
        remove = np.random.randint(2, size=6)
        remove[0] = 0
        c = 0
        change_count = [0]
        hand_type = [0]*10
        stuck = False

        for i_ in remove:
            if i_ == 1:
                new_h = pe_.generate_hand(1)

                if new_h[0] in _hand:
                    stuck = True

                while stuck:
                    new_h = pe_.generate_hand(1)
                    if new_h[0] not in _hand:
                        stuck = False

                _hand[c] = copy.deepcopy(new_h[0])
                change_count[0] += 1
            c += 1
        """
        Get the new hand info
        """
        _hand = _hand[1:]
        _hand = card_sorter(_hand)
        new_hand_strengthx = evaluate_hand(_hand)

        serialized = serialize(_start_hand)
        s = [i * hand_strengthx for i in serialized]
        diff = (new_hand_strengthx-hand_strengthx)
        data = s + RemovedCard[change_count[0]]
        X.append(data)
        Y.append(diff)
        count += 1
        if (count%5000) == 0:
            print "training [" + coolstuff[funny] + "]"
            funny += 1
            if funny == 3: funny = 0
            count = 0

    return X,Y


if __name__ == "__main__":

    train = False

    filename = '10mil.sav'
    r = GradientBoostingRegressor()
    # TRAIN
    if train == True:
        x,y = gen_dataset(10000000)
        r.fit(x,y)
        joblib.dump(r, filename)
        print "Training done!"
    # TEST TRAIN MODEL
    else:
        # LOAD MODEL
        loaded_model = joblib.load(filename)
        # GENEREATE AND SORT HAND
        _hand = pe_.generate_hand()
        _sorted_hand = card_sorter(_hand)
        # GET VALUE OF HAND
        hand_strengthx = evaluate_hand(_sorted_hand)
        # SERIALZE HAND
        serialized = serialize(_sorted_hand)
        s = [i * hand_strengthx for i in serialized]
        # CHECK ALL POSIBLE ACTIONS
        t = time.time()
        allFeatures = []
        for a_ in range(0,6): allFeatures.append(s+RemovedCard[a_])
        # PREDICT BEST ACTION
        print _sorted_hand
        pReturn = loaded_model.predict(allFeatures)
        stop = time.time()
        print pReturn
        # ACTION CHOSEN
        action = ChangeAction[np.argmax(pReturn)]
        print "Prediction time (%.5f)" % (stop-t)
        print action
