# Percepton:

# input xn: x1+x2+...+xn>THRESHOLD

# Weight w1, w2,..., wn decides the importance of respective input; e.g w1 = 2, w2 = 30, w3 = 1
# x1*w1+x2*w2+x3*w3 > THRESHOLD. If the value is greater than the THRESHOLD it will produce the
# output 1, else 0.

# If there is 2 inputs, you need however 3 weights. One for each
# input and one for a constant bias b -> b*w3

# Generate a dataset and plot it

import numpy as np

# If the output range is 1, the network produces a value between 0 and 1. Range 1000
# produces a value between 0 and 1000.
output_range = 1000

# Output values are the desired output you want to teach the network.
# Output values:
a = 100
b = 200
c = 300
d = 400
e = 500
f = 300
# The output values are divided by the output range to do calculation with the sigmoid
# activation function between 0 and 1. The system later have to multiply the output weights
# with the output range.

# Sigmoid function, activation function?
# Sigmoid function is a function that produces a S-shaped curve.
def nonlin(x, deriv = False):
    if deriv:
        return x*(1-x)
    return (1/(1+np.exp(-x)))
# If nonlin is called with (x, True), it will instead of the value at given point
# return the slope, the derivative
# If the sigmoid function produces the value out "out", then the derivative is out*(1-out)?

# Choose a dataset.
# Import dataset.
dataset_in = np.array([[1, 0, 1],[0,0,1],
                       [1, 0, 1],[1,0,1],
                       [0, 0, 1],[1,0,1],
                       [0, 0, 1],[0,0,1],
                       [0, 0, 1],[1,0,1],
                       [0, 0, 1],[1,0,1],
                       [1, 0, 1],[0,0,1],
                       [1, 0, 1],[0,0,1]
                        ])

# Input data is know.

# Output dataset for the input, also known in this case.

dataset_out = np.array([[1],[0],
                        [1],[1],
                        [0],[1],
                        [0],[0],
                        [0],[1],
                        [0],[1],
                        [1],[0],
                        [1],[0]
                        ])

# seed random numbers to make calculation
# deterministic (just a good practice)
np.random.seed(1)


# initialize weights randomly with mean 0
# The weight layer 0 have the matrix dimension 3x4 because there is 3 inputs
# and 4 outputs (1 for each to the next weight layer
# The weight layer 1 therefor have the dimension 4x1. 4 inputs from the
# first weight layer and 1 for the output
weight_layer_0 = syn0 = 2*np.random.random((len(dataset_in[0]),7)) - 1
weight_layer_1 = syn1 = 2*np.random.random((7,6)) - 1
weight_layer_2 = syn2 = 2*np.random.random((6, len(dataset_out[0]))) - 1
# The weight layer 1 connects the input layer with the first hidden layer.

for i in range(50000):        # A very high training iteration seem to produce more accurate predictions.
                                # It is worth noting that a too high iteration number may either produce,
                                # worse prediction(Unlikely) or same result, which means a unnecassary amount
                                # of iterations.

    # Forward propagation
    input_layer     = l0 = dataset_in
    hidden_layer_1  = l1 = nonlin(np.dot(input_layer, weight_layer_0))
    hidden_layer_2  = l2 = nonlin(np.dot(hidden_layer_1, weight_layer_1))
    hidden_layer_3  = l3 = nonlin(np.dot(hidden_layer_2, weight_layer_2))

    # With how much did we miss?

    layer_3_error = l3_error = dataset_out - hidden_layer_3

    if(i%10000) == 0:
        print("Error: " + str(np.mean(np.abs(layer_3_error))))



    # Multiply how much we missed
    # by the slope of the activation function, sigmoid, at the values in hidden_layer_2
    hidden_layer_3_Delta = layer_3_error * nonlin(hidden_layer_3, deriv = True)

    layer_2_error = hidden_layer_3_Delta.dot(weight_layer_2.T)

    hidden_layer_2_Delta = layer_2_error * nonlin(hidden_layer_2, deriv = True)

    layer_1_error = hidden_layer_2_Delta.dot(weight_layer_1.T)

    hidden_layer_1_Delta = layer_1_error * nonlin(hidden_layer_1, deriv=True)


    weight_layer_2 += hidden_layer_2.T.dot(hidden_layer_3_Delta)
    weight_layer_1 += hidden_layer_1.T.dot(hidden_layer_2_Delta)
    weight_layer_0 += input_layer.T.dot(hidden_layer_1_Delta)

print("Output after training")
print(hidden_layer_3*output_range)
print("\n")


#Used too test combinations and if the system is trained well, and/or as pleased.
test_layer      = np.array([[1, 0, 1]])
test_hidden_layer_1 = nonlin(np.dot(test_layer, weight_layer_0))
test_hidden_layer_2 = nonlin(np.dot(test_hidden_layer_1, weight_layer_1))
test_hidden_layer_3 = nonlin(np.dot(test_hidden_layer_2, weight_layer_2))
print(test_hidden_layer_3)

test_layer      = np.array([[1, 0, 1]])
test_hidden_layer_1 = nonlin(np.dot(test_layer, weight_layer_0))
test_hidden_layer_2 = nonlin(np.dot(test_hidden_layer_1, weight_layer_1))
test_hidden_layer_3 = nonlin(np.dot(test_hidden_layer_2, weight_layer_2))
print(test_hidden_layer_3)

test_layer      = np.array([[0, 0, 1]])
test_hidden_layer_1 = nonlin(np.dot(test_layer, weight_layer_0))
test_hidden_layer_2 = nonlin(np.dot(test_hidden_layer_1, weight_layer_1))
test_hidden_layer_3 = nonlin(np.dot(test_hidden_layer_2, weight_layer_2))
print(test_hidden_layer_3)